unit Proxy;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,TypInfo, Buttons, DsgnIntf,ExtCtrls, Menus;

  {$Include handel.inc}

type
  TGrabPosition = (gpBottomLeft, gpLeft, gpTopLeft, gpTop,
                   gpTopRight, gpRight, gpBottomRight, gpBottom);
  TGrabHandles = class;
  TGrabHandle = class(TCustomControl)
  private
    fPosition: TGrabPosition;
    fControl: TControl;
    fDragging: Boolean;
    fDragPoint: TPoint;
    fDragRect: TRect;
    fSize: Cardinal;
    fHandles: TGrabHandles;
    fColor:TColor;
    fMultiSelected:Boolean;
    procedure SetMultiSelected(Value:Boolean);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure StartDrag(X, Y: Integer);
    procedure DoDrag(X, Y: Integer);
    procedure EndDrag(X, Y: Integer);
    procedure ChooseCursor;
  public
    constructor Create(Control: TControl; Position: TGrabPosition; Handles: TGrabHandles);
    procedure Paint; override;
    procedure SetGrabBounds;
    function Center: TPoint;
    property Color:TColor read fColor write fColor default clWhite;
    property GrabPosition: TGrabPosition read fPosition;
    property Control: TControl read fControl;
    property DraggingTo: Boolean read fDragging;
    property DragPoint: TPoint read fDragPoint;
    property DragRect: TRect read fDragRect;
    property GrabHandles: TGrabHandles read fHandles;
    property MultiSelected:Boolean read fMultiSelected write SetMultiSelected;
    property Size: Cardinal read fSize;
  end;

  { Array of grab handles at strategic locations around a component. }
  TGrabHandleArray = array[Low(TGrabPosition)..High(TGrabPosition)] of TGrabHandle;
  TGrabHandles = class
  private
    fHandles: TGrabHandleArray;
    fVisible: Boolean;
    fColor:TColor;
    fControl:TControl;
    fMultiSelected:Boolean;
    function GetHandle(Index: TGrabPosition): TGrabHandle;
    procedure SetVisible(Value: Boolean);
    procedure SetColor(Value:TColor);
    procedure SetMultiSelected(Value:Boolean);
  public
    constructor Create(Control: TControl);
    destructor Destroy; override;
    property  Color:TColor read fColor write SetColor;
    property  Handle[Index: TGrabPosition]: TGrabHandle read GetHandle;
    procedure Hide;
    procedure Show;
    procedure Update;
    property  Control:TControl read fControl write fControl;
    property  MultiSelected:Boolean read fMultiSelected write SetMultiSelected;
    property  Visible: Boolean read fVisible write SetVisible;
  end;

  TDragRectArray = array [1..255] of TRect;

  TDragRectItem = class
  private
     fRectArray: TDragRectArray;
     function GetItem(Index: Integer): TRect;
     procedure SetItem(Index:Integer;Value:TRect);
  public
     constructor Create;
     procedure Clear;
     property Item[Index:Integer]: TRect read GetItem write SetItem ;
  end;

  TDragRectList = class
  private
     fCount: Integer;
     fRectItem: TDragRectItem;
  public
     constructor Create;
     procedure Add(Control:TControl);
     procedure Clear;
     property Items: TDragRectItem read fRectItem write fRectItem ;
  end;

  { A control wrapper for non-visual components. }
  TWrapperControl = class(TCustomControl)
  private
    fComponent: TComponent;
    fBitmap: TBitmap;
  protected
    procedure MakeBitmap;
  public
    constructor Create(Owner: TComponent; Component: TComponent);
    destructor Destroy; override;
    procedure Paint; override;
    procedure UpdateControl;
    property  Component: TComponent read fComponent;
    property  Bitmap: TBitmap read fBitmap;
  published
    property  OriginComponent:TComponent read fComponent write fComponent;
  end;

  TProxyForm = class(TForm)
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    fControl: TControl;
    fDragging: Boolean;
    fDragPoint: TPoint;
    fDragRect: TRect;
    fGrabHandles: TGrabHandles;
    fComponentList: TStringList;
    fDragRectList: TDragRectList;
    GotMouse : boolean;
    Anchor, Rover : TPoint;
    procedure MakeARubber(X, Y : integer);
    procedure DrawRect(const Rect: TRect);
    procedure SaveTempForm;
  protected
    procedure ClipChildren(Clipping: Boolean);
    procedure SelectComponentsInRect(Rect: TRect);
    property  DraggingTo: Boolean read fDragging;
    property  DragPoint: TPoint read fDragPoint;
    property  DragRect: TRect read fDragRect;
  public
    { Public declarations }
    FileName:string;
    procedure StartDragging(Pt: TPoint);
    procedure EndDragging(Pt: TPoint);
    procedure DragTo(Pt: TPoint);
    procedure ClipCursorToComponentArea;
    procedure SetCompBounds(const Bounds: TRect);
    function  GetCompBounds: TRect;
    procedure OnMenuClick(Sender:TObject);
    procedure OnFindMethodHandler(Reader: TReader; const MethodName: string;
              var Address: Pointer; var Error: Boolean);
    procedure OnReaderErrorHandler(Reader: TReader; const Message: string; var Handled: Boolean);
    property  ComponentList:TStringList read fComponentList write fComponentList;
    property  DragRectList: TDragRectList read fDragRectList write fDragRectList;
    // Control Ӽ   Ʈ   ִ´.
    property  SelectControl: TControl read fControl write fControl;
    property  GrabHandles: TGrabHandles read fGrabHandles write fGrabHandles;
  end;

var
  ProxyForm: TProxyForm;

implementation

uses ObjectInspec, MainForm, Uconst,utype, Editor;

{$R *.DFM}
// ð ۳Ʈ Ʈ   Ʒ ۳Ʈ ҽ ʿϴ.
{$R STDREG.DCR}
{$R SYSREG.DCR}
{$R DBREG.DCR}

// 콺 ġ 巡׿ شϴ Ÿŭ ° Ѵ.
function IsMouseDrag(Old, New: TPoint): Boolean;
var
  DifX, DifY:Integer;
begin
   Result:= False;
   DifX  := Abs(New.x - Old.x);
   DifY  := Abs(New.y - Old.y);
   if (DifX > 5) or (DifY > 5) then Result:= True
   else Result:= False;
end;

{ Create a grab handle at a specific position, for a control. }
constructor TGrabHandle.Create(Control: TControl; Position: TGrabPosition; Handles: TGrabHandles);
begin
  inherited Create(Control.Owner);
  ControlStyle := ControlStyle - [csOpaque];
  Parent       := Control.Parent;
  fColor       := clWhite;
  fControl     := Control;
  fHandles     := Handles;
  fPosition    := Position;
  fSize        := Screen.PixelsPerInch div 32;
  // exclude csDesigning flag from grab control's componentstate for sizing control 
  TExposeComponent(self).SetDesigning(False);
  SetGrabBounds;
  ChooseCursor;
end;

{ Return the center coordinates of the grab handle. }
function TGrabHandle.Center: TPoint;
begin
  case GrabPosition of
    gpTopLeft:
       if fMultiSelected then Center := Point(Control.Left+2, Control.Top+2)
       else Center := Point(Control.Left, Control.Top);
    gpTop:
       Center := Point(Control.Left + Control.Width div 2, Control.Top);
    gpTopRight:
       if fMultiSelected then Center := Point(Control.Left + Control.Width -2, Control.Top+2)
       else Center := Point(Control.Left + Control.Width, Control.Top);
    gpRight:
       if fMultiSelected then
            Center := Point(Control.Left + Control.Width - 2, Control.Top + Control.Height div 2)
       else Center := Point(Control.Left + Control.Width, Control.Top + Control.Height div 2);
    gpBottomRight:
       if fMultiSelected then Center := Point(Control.Left + Control.Width-2, Control.Top + Control.Height-2)
       else Center := Point(Control.Left + Control.Width, Control.Top + Control.Height);
    gpBottom:
       Center := Point(Control.Left + Control.Width div 2, Control.Top + Control.Height);
    gpBottomLeft:
       if fMultiSelected then Center := Point(Control.Left+2, Control.Top + Control.Height-2)
       else Center := Point(Control.Left, Control.Top + Control.Height);
    gpLeft:
       if fMultiSelected then
            Center := Point(Control.Left + 2, Control.Top + Control.Height div 2)
       else Center := Point(Control.Left, Control.Top + Control.Height div 2);
  end;
end;

{ Set the cursor, depending on the position of the handle. }
procedure TGrabHandle.ChooseCursor;
begin
  case GrabPosition of
    gpTopLeft, gpBottomRight: Cursor := crSizeNWSE;
    gpTop, gpBottom:          Cursor := crSizeNS;
    gpTopRight, gpBottomLeft: Cursor := crSizeNESW;
    gpRight, gpLeft:          Cursor := crSizeWE;
  end;
end;

{ Set the boundaries of the grab handle. }
procedure TGrabHandle.SetGrabBounds;
begin
  with Center do
    inherited SetBounds(X - LongInt(Size), Y - LongInt(Size), Size*2, Size*2);
  Invalidate;
end;

procedure TGrabHandle.Paint;
begin
  if GrabHandles.Visible then
  begin
    Canvas.Brush.Color := FColor;
    Canvas.Brush.Style := bsSolid;
   // Canvas.FillRect(ClientRect);
    Canvas.Rectangle(ClientRect.Left, ClientRect.Top, ClientRect.Right, ClientRect.Bottom);
  end;
end;

{ Left button down on a grab handle means the user is resizing the control. }
procedure TGrabHandle.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then StartDrag(X, Y);
end;

{ While resizing, drag the sizing rectangle. }
procedure TGrabHandle.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if DraggingTo then  DoDrag(X, Y);
end;

{ Mouse up: stop dragging. }
procedure TGrabHandle.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if DraggingTo then  EndDrag(X, Y);
end;

{ Start dragging to resize the control. Hide the handles, and
  show the sizing rectangle. Restrict the cursor to the editing
  area, so the user doesn't move nto the method list. }
procedure TGrabHandle.StartDrag(X, Y: Integer);
var
  ControlPt: TPoint;
begin
  fDragging  := True;
  ControlPt  := Center;
  fDragPoint := Point(X - ControlPt.X, Y - ControlPt.Y);
  GrabHandles.Hide;

  fDragRect := Control.BoundsRect;
  with Owner as TProxyForm do
  begin
    ClipCursorToComponentArea;
    ClipChildren(False);
    DrawRect(Self.DragRect);
  end;
end;

{ Continue dragging the sizing rectangle. If the user drags the corner
  across the control, the corners might need to be swapped. }
procedure TGrabHandle.DoDrag(X, Y: Integer);
  procedure Swap(var A, B: Integer);
  var
    Tmp: Integer;
  begin
    Tmp := A;
    A   := B;
    B   := Tmp;
  end;
var
  OldRect, NewRect: TRect;
begin
  with Owner as TProxyForm do
   DrawRect(Self.DragRect);
  X := X - DragPoint.X;
  Y := Y - DragPoint.Y;
  OldRect := Control.BoundsRect;
  case GrabPosition of
  gpTopLeft:     NewRect := Rect(X, Y, OldRect.Right, OldRect.Bottom);
  gpTop:         NewRect := Rect(OldRect.Left, Y, OldRect.Right, OldRect.Bottom);
  gpTopRight:    NewRect := Rect(OldRect.Left, Y, X, OldRect.Bottom);
  gpRight:       NewRect := Rect(OldRect.Left, OldRect.Top, X, OldRect.Bottom);
  gpBottomRight: NewRect := Rect(OldRect.Left, OldRect.Top, X, Y);
  gpBottom:      NewRect := Rect(OldRect.Left, OldRect.Top, OldRect.Right, Y);
  gpBottomLeft:  NewRect := Rect(X, OldRect.Top, OldRect.Right, Y);
  gpLeft:        NewRect := Rect(X, OldRect.Top, OldRect.Right, OldRect.Bottom);
  end;
  with NewRect do
  begin
    if Top > Bottom then
      Swap(Top, Bottom);
    if Left > Right then
      Swap(Left, Right);
  end;
  fDragRect := NewRect;
  with Owner as TProxyForm do
    DrawRect(Self.DragRect);
end;

{ Stop dragging the sizing rectangle. }
procedure TGrabHandle.EndDrag(X, Y: Integer);
var
  Rect: TRect;
begin
  with Owner as TProxyForm do
  begin
    ClipChildren(True);
    DrawRect(Self.DragRect);
  end;
  fDragging := False;
  ClipCursor(nil);

  { Some components are fixed size. If so, keep the origin,
    but reset the size to the fixed size. }
  Rect := DragRect;
  if csFixedWidth in Control.ControlStyle then
    Rect.Right := Rect.Left + Control.Width;
  if csFixedHeight in Control.ControlStyle then
    Rect.Bottom := Rect.Top + Control.Height;
  with Rect do
    Control.SetBounds(Left, Top, Right-Left, Bottom-Top);
  GrabHandles.Show;
  ObjectInspector.DisplayProperty(Control,Control);  // Update Properties
end;

procedure TGrabHandle.SetMultiSelected(Value:Boolean);
begin
  if fMultiSelected = Value then Exit;
  fMultiSelected:= Value;
end;

{TGrabHandles}

{ Create a set of grab handles, at the corners and sides of a control. }
constructor TGrabHandles.Create(Control: TControl);
var
  Pos: TGrabPosition;
begin
  inherited Create;
  fVisible := True;
  fColor   := clWhite;
  fControl := Control;
  fMultiSelected:= False;
  for Pos := Low(TGrabPosition) to High(TGrabPosition) do
    fHandles[Pos] := TGrabHandle.Create(Control, Pos, Self);
end;

destructor TGrabHandles.Destroy;
var
  Pos: TGrabPosition;
begin
  for Pos := Low(TGrabPosition) to High(TGrabPosition) do
    fHandles[Pos].Free;
  inherited Destroy;
end;

procedure TGrabHandles.SetColor(Value:TColor);
begin
   if fColor = Value then Exit;
   fColor:= Value;
   Update;
end;

{ Return a specific handle. }
function TGrabHandles.GetHandle(Index: TGrabPosition): TGrabHandle;
begin
  Result := fHandles[Index];
end;

{ Hide all the grab handles, when dragging. }
procedure TGrabHandles.Hide;
var
  Pos: TGrabPosition;
begin
  if Visible then
  begin
    fVisible := False;
    for Pos := Low(TGrabPosition) to High(TGrabPosition) do
      fHandles[Pos].Hide;
  end;
end;

{ Show all the grab handles again. }
procedure TGrabHandles.Show;
var
  Pos: TGrabPosition;
begin
  if not Visible then
  begin
    fVisible := True;
    for Pos := Low(TGrabPosition) to High(TGrabPosition) do
    begin
      if fHandles[Pos] = nil then Exit;
      fHandles[Pos].SetGrabBounds;
      fHandles[Pos].Color:= fColor;
      fHandles[Pos].Show;
    end;
  end;
end;

{ Update the position of the grab handles after resizing or moving. }
procedure TGrabHandles.Update;
var
  Pos: TGrabPosition;
begin
  if Visible then
    for Pos := Low(TGrabPosition) to High(TGrabPosition) do
    begin
      fHandles[Pos].SetGrabBounds;
      fHandles[Pos].Color:= fColor;
    end;
end;

{ Set the visibility of the grab handles. }
procedure TGrabHandles.SetVisible(Value: Boolean);
begin
  if Value then Show
  else  Hide
end;

//  ۳Ʈ ÿ ϴ 쿡
// GrabHandle ġ ,Visible θ Ѵ.
procedure TGrabHandles.SetMultiSelected(Value:Boolean);
var
  Pos: TGrabPosition;
begin
   if fMultiSelected = Value then Exit;
   fMultiSelected:= Value;
   if fMultiSelected then
   begin
      for Pos := Low(TGrabPosition) to High(TGrabPosition) do
      begin
       { if not (Pos in [gpBottomLeft,  gpTopLeft, gpTopRight,  gpBottomRight]) then
           fHandles[Pos].Visible:= False
        else}
        begin
           fHandles[Pos].MultiSelected:= True;
           fHandles[Pos].Color:= fColor;
           fHandles[Pos].SetGrabBounds;
           fHandles[Pos].Show;
        end;
      end;
   end;
end;

{TDragRectItem}
constructor TDragRectItem.Create;
begin
   Clear;
end;

procedure TDragRectItem.Clear;
var
  I:Integer;
begin
   for I:= Low(TDragRectArray) to High(TDragRectArray) do
      fRectArray[I].Left:= -1;
end;

function TDragRectItem.GetItem(Index: Integer): TRect;
begin
   Result:= fRectArray[Index];
end;

procedure TDragRectItem.SetItem(Index:Integer; Value:TRect);
begin
   fRectArray[Index]:= Value;
end;

{TDragRectList}

{TDragRectList  ۳Ʈ   巡 
 巡׵Ǵ ۳Ʈ ġ ֱ  Ѵ.}
constructor TDragRectList.Create;
begin
   fRectItem:= TDragRectItem.Create;
   Clear;
end;

procedure TDragRectList.Add(Control:TControl);
begin
   if fCount >= High(TDragRectArray) then Exit;
   fCount:= fCount + 1;
   fRectItem.Item[fCount]:= Control.BoundsRect;
end;

procedure TDragRectList.Clear;
begin
   fRectItem.Clear;
   fCount:=0;
end;

{TWrapperControl}

{ The wrapper control is used for non-visual components. The component
  is drawn as a button, so the user can move it, double click on it, etc. }
constructor TWrapperControl.Create(Owner: TComponent; Component: TComponent);
begin
  inherited Create(Owner);
  Parent       := Owner as TWinControl;
  ControlStyle := ControlStyle + [csFixedWidth, csFixedHeight];
  fComponent   := Component;
  fBitmap      := TBitmap.Create;
  Bitmap.Height:= 28;
  Bitmap.Width := 28;
  Height       := Bitmap.Height;
  Width        := Bitmap.Width;
  MakeBitmap;
end;

destructor TWrapperControl.Destroy;
begin
  Bitmap.Free;
  inherited Destroy;
end;

{ Make the button bitmap by looking up its bitmap resource, which
  was in its .DCR file. }
procedure TWrapperControl.MakeBitmap;
var
  ResName: array[0..64] of Char;
  ResBitmap: TBitmap;
  R: TRect;
  X, Y: Integer;
begin
  ResBitmap := TBitmap.Create;
  try
    StrPLCopy(ResName, Component.ClassName, SizeOf(ResName));
    AnsiUpper(ResName);
    ResBitmap.Handle := LoadBitmap(hInstance, ResName);
    if ResBitmap.Handle = 0 then
      ResBitmap.Handle := LoadBitmap(hInstance, 'DEFAULT');

    Bitmap.Canvas.Brush.Color := clBtnFace;
    Bitmap.Canvas.Pen.Color := clWindowText;
    Bitmap.Canvas.Rectangle(0, 0, Bitmap.Width, Bitmap.Height);
    R := Rect(1, 1, Bitmap.Width-1, Bitmap.Height-1);
    Frame3D(Bitmap.Canvas, R, clBtnHighlight, clBtnShadow, 2);

    X := (Bitmap.Width - ResBitmap.Width) div 2;
    Y := (Bitmap.Height - ResBitmap.Height) div 2;
    Bitmap.Canvas.BrushCopy(Bounds(X, Y, ResBitmap.Width, ResBitmap.Height),
                            ResBitmap,
                            Rect(0, 0, ResBitmap.Width, ResBitmap.Height),
                            ResBitmap.TransparentColor);
  finally
    ResBitmap.Free;
  end;
end;

{ Paint the button. }
procedure TWrapperControl.Paint;
var
  X, Y: Integer;
begin
  if Bitmap = nil then
  begin
  //   д   Ʈ Ʈ Ʈ ش.
    fBitmap := TBitmap.Create;
    Bitmap.Height := 28;
    Bitmap.Width := 28;
    Bitmap.Handle := LoadBitmap(hInstance, 'DEFAULT');
  end;
  X := (ClientWidth - Bitmap.Width) div 2;
  Y := (ClientHeight - Bitmap.Height) div 2;
  Canvas.Draw(X, Y, Bitmap);
end;

procedure TWrapperControl.UpdateControl;
begin
  if Bitmap = nil then
  begin
      fBitmap       := TBitmap.Create;
      Bitmap.Height := 28;
      Bitmap.Width  := 28;
      fComponent    := OriginComponent;
      if fComponent <> nil then MakeBitmap
      else Bitmap.Handle := LoadBitmap(hInstance, 'DEFAULT');
  end;
end;

{ ۳Ʈ ComponentState csDesigning   Ŭ.
 SetDesiging ޼尡 protecteḏ  public ٲٱ 
 ο Ŭ . SetDesigning  True ϸ
 ComponentState Ӽ csDesigning Եȴ.}
type
  TExposeComponent = class(TComponent)
  public
    procedure SetDesigning(Value: Boolean);
  end;

procedure TExposeComponent.SetDesigning(Value: Boolean);
begin
  inherited SetDesigning(Value);
end;

{TProxyForm}

{ Set the component's boundaries. }
procedure TProxyForm.SetCompBounds(const Bounds: TRect);
begin
  with Bounds do
    SelectControl.SetBounds(Left, Top, Right-Left, Bottom-Top);
end;

{ Return the component's boundaries. }
function TProxyForm.GetCompBounds: TRect;
begin
  Result := SelectControl.BoundsRect;
end;

{ Set the cursor clipping region to the component editing area.
  This prevents the user from dragging the component into the
  method list or status bar. }
procedure TProxyForm.ClipCursorToComponentArea;
var
  Rect: TRect;
begin
  Rect.TopLeft := ClientToScreen(Point(0, 0));
  Rect.BottomRight := ClientToScreen(Point(Width, ClientHeight));
  ClipCursor(@Rect);
end;

{ Draw the component's frame while moving it. }
procedure TProxyForm.DrawRect(const Rect: TRect);
begin
  with Canvas do
  begin
     Pen.Color   := Color;
     Pen.Mode    := pmXor;
     Pen.Style   := psDot;
     Pen.Width   := 2;
     Brush.Style := bsClear;
  end;
  with Rect do
    Canvas.Rectangle(Left, Top, Right, Bottom);
  fDragRect := Rect;
end;

{ Start moving the component by hiding its grab handles and
  drawing the frame to represent the component. }
procedure TProxyForm.StartDragging(Pt: TPoint);
var
  I:Integer;
  R:TRect;
begin
  if ProxyDesigner =  nil then Exit;
  if ProxyDesigner.MultiSelected then Exit;
  if not IsMouseDrag(fDragPoint, Pt) then Exit;
  fDragging := True;
  fDragPoint := Pt;
  ClipChildren(False);
  for I:= Low(TDragRectArray) to High(TDragRectArray) do
  begin
     if DragRectList.Items.Item[I].Left = -1 then Break;
     R := DragRectList.Items.Item[I];
     DrawRect(R);
  end;
  ClipCursorToComponentArea;
end;

{ Draw the component's frame and redraw it. }
procedure TProxyForm.DragTo(Pt: TPoint);
var
  R: TRect;
  I:Integer;
  Control:TControl;
begin
  if ProxyDesigner =  nil then Exit;
  for I:=Low(TDragRectArray) to High(TDragRectArray) do
  begin
     if DragRectList.Items.Item[I].Left = -1 then Break;
     Control:= TControl(ProxyDesigner.SelectList[I-1]);
     DrawRect(DragRectList.Items.Item[I]);
     R        := DragRectList.Items.Item[I];
     R.Left   := R.Left + Pt.X - DragPoint.X;
     R.Top    := R.Top  + Pt.Y - DragPoint.Y;
     R.Right  := R.Left + Control.Width;
     R.Bottom := R.Top  + Control.Height;
     DrawRect(R);
     DragRectList.Items.Item[I]:= R;
  end;
  fDragPoint := Pt;
end;

// ۳Ʈ 巡׸  , ۳Ʈ ġ Ѵ.
procedure TProxyForm.EndDragging(Pt: TPoint);
var
  I:Integer;
  Control:TControl;
  R:TRect;
begin
  ClipChildren(True);
  if (Pt.X <> DragPoint.X) or (Pt.Y <> DragPoint.Y) then DragTo(Pt);
  for I:= Low(TDragRectArray) to High(TDragRectArray) do
  begin
     if DragRectList.Items.Item[I].Left = -1 then Break;
     Control:= TControl(ProxyDesigner.SelectList[I-1]);
     R := DragRectList.Items.Item[I];
     DrawRect(R);
     Control.Left := R.Left;
     Control.Top  := R.Top;
  end;
  fDragging := False;
  ClipCursor(nil);
  if not (SelectControl is TWrapperControl) then
    if ProxyDesigner.SelectList.Count = 1 then
       ObjectInspector.DisplayProperty(SelectControl,SelectControl); // Update Properties
  ProxyDesigner.UpdateGrabHandle;
  ProxyDesigner.PaintGrid;
end;

{ Turn on or off clipping for child components. When dragging
  the separator panel or the component, turning off the clipping
  draws the highlight over all controls. }
procedure TProxyForm.ClipChildren(Clipping: Boolean);
var
  Style: LongInt;
begin
  Style := GetWindowLong(Handle, Gwl_Style);
  if Clipping then
    Style := Style or Ws_ClipChildren
  else
    Style := Style and not Ws_ClipChildren;
  SetWindowLong(Handle, Gwl_Style, Style);
end;

//  콺 Ŭ  ο ۳Ʈ ؾ ϴ
// 巡 θ Ѵ.
procedure TProxyForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  Pt: TPoint;
  SelClass:TComponentClass;
begin
   if ProxyDesigner = nil then Exit;
   SelClass:= FMainForm.GetSelectComponent;
   if SelClass <> nil then
   begin
      if (SelectControl <> nil) and (SelectControl is TCustomControl) then
         ProxyDesigner.CreateComponent(SelClass, SelectControl, X, Y, 0, 0)
      else ProxyDesigner.CreateComponent(SelClass, ProxyDesigner.Form, X, Y, 0, 0);
      Exit;
   end;
   if (SelectControl = nil) then Exit;
   Pt := Point(X, Y);
   Pt := Self.ScreenToClient(SelectControl.ClientToScreen(Pt));
   if PtInRect(GetCompBounds, Pt) then  StartDragging(Pt)
   else
   begin
      ObjectInspector.DisplayProperty(nil,self);
      SetCapture(Handle);  { WIN API function, grabs all mouse actions to this window }
      GotMouse := True;    { need to keep track of who has teh mouse }
      Anchor.X := X; Anchor.Y := Y;  { where we started from }
      Rover := Anchor;               { where we are now}
      Canvas.MoveTo(X,Y);
   end;
end;

procedure TProxyForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  P:TPoint;
begin
  if ProxyDesigner = nil then Exit;
  // Shift Ű ۳Ʈ ϰ ִ 쿡 巡׸  ʴ´.
  if ProxyDesigner.MultiSelected then Exit;
  P:= Point(X, Y);
  // Ư Ÿŭ 콺 ̵ؾ 巡ϴ  Ѵ.
  if not IsMouseDrag(fDragPoint, P) then Exit;
  if DraggingTo then DragTo(Point(X, Y));
  if GotMouse then MakeARubber(X,Y);
end;

procedure TProxyForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
   if ProxyDesigner = nil then Exit;
   if DraggingTo then
   begin
      EndDragging(Point(X, Y));
      FMainForm.UpdateControl(psChange);
      Exit;
   end;
   if GotMouse then
   begin
     with Canvas do
     begin
       SetROP2(Handle,R2_NOTXORPEN);
       Pen.Style   := psDot;
       Brush.Style := bsClear;
       Rectangle(Anchor.X,Anchor.Y,Rover.X,Rover.Y);
     end;
     ReleaseCapture;
     SelectComponentsInRect(Rect(Anchor.X,Anchor.Y,Rover.X,Rover.Y));
     GotMouse := False;
   end;
end;

procedure TProxyForm.FormCreate(Sender: TObject);
begin
   {$IFDEF DELPHI3}
     Designer := ProxyDesigner;
   {$ENDIF}
   {$IFDEF DELPHI4}
     Designer := IFormDesigner(ProxyDesigner);
   {$ENDIF}
   fComponentList:= TStringList.Create;
   fDragRectList := TDragRectList.Create;
  // ProxyDesigner.SetProxyForm(self);
end;

procedure TProxyForm.FormActivate(Sender: TObject);
begin
   FMainForm.ActiveForm:= self;
   FMainForm.ToggleFormUnit:= True;
   ProxyDesigner.SetProxyForm(self);
end;

procedure TProxyForm.SaveTempForm;
var
  Output:TMemoryStream;
  I:Integer;
begin
    Output:=TMemoryStream.Create;
    for I:= ComponentCount-1 downto 0 do
    begin
       if Components[I] is TGrabHandle then  Components[I].Free
       else if Components[I].Name = 'ProxyPopupMenu1' then Components[I].Free;
    end;
    OutPut.WriteComponentRes(FilePath + TempForm,self);
    Output.SaveToFile(FilePath+TempForm);
    Output.Free;
end;

// Event handler for component's popup menu
procedure TProxyForm.OnMenuClick(Sender:TObject);
var
  Name:string;
begin
   Name:= TMenuItem(Sender).Name;
   if      Name = 'pmFront' then SelectControl.BringToFront
   else if Name = 'pmBack'  then SelectControl.SendToBack
   else if Name = 'pmCopy'  then ProxyDesigner.CopyComponent
   else if Name = 'pmCut'   then ProxyDesigner.CutComponent
   else if Name = 'pmPaste' then ProxyDesigner.PasteComponent
   else if Name = 'pmSize'  then FMainForm.EditSizeItemClick(self)
   else if Name = 'pmScale' then FMainForm.EditScaleItemClick(self)
   else if Name = 'pmAlign' then FMainForm.EditAlignItemClick(self)
   else if Name = 'pmViewAsText' then
   begin
      SaveTempForm;
      FMainForm.OpenFormAsText(FilePath+TempForm);
      EditorForm.PageControl1.ActivePage.Caption:= FileName;
   end
   else if Name = 'pmDelete' then ProxyDesigner.DeleteSelectList ;
end;

procedure TProxyForm.FormDestroy(Sender: TObject);
begin
   fComponentList.Free;
   fDragRectList.Free;
end;

procedure TProxyForm.MakeARubber(X, Y : integer);
begin
  with Canvas do
  begin
    SetROP2(Handle,R2_NOTXORPEN);  { use to Raster Op codes to make the rubberband }
    Pen.Style   := psDot;
    Brush.Style := bsClear;        { don't fill the interior of the shape, please }
    Rectangle(Anchor.X,Anchor.Y,Rover.X,Rover.Y);
    Rover.X := X;
    Rover.Y := Y;
    Rectangle(Anchor.X,Anchor.Y,Rover.X,Rover.Y);
  end;
end;

procedure TProxyForm.SelectComponentsInRect(Rect: TRect);
// Select components in Rect
var
  I: Integer;
  Control: TControl;
  rec: TRect;
  List: TComponentList;
  rgn: HRGN;
begin
   rgn:= CreateRectRgn(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
   List:= TComponentList.Create;
   for I:= 0 to ComponentCount - 1 do
   begin
      Control:= TControl(Components[I]);
      rec:= Control.BoundsRect;
      if RectInRegion(rgn, rec) then List.Add(Control);
   end;
   {$IFDEF DELPHI4UP}
   //ProxyDesigner.SetSelections(List);
   ProxyDesigner.SetSelectionList(List);
   {$ENDIF}
   List.Free;
end;

procedure TProxyForm.OnFindMethodHandler(Reader: TReader; const MethodName: string;
    var Address: Pointer; var Error: Boolean);
begin
    Address:= ProxyDesigner.Form.MethodAddress(MethodName);
    if Address <> nil then
       ProxyDesigner.Methods[Integer(Address)]:= MethodName
    else
    begin
       ProxyDesigner.Methods.AddObject(MethodName, Address);
       Error:= False;
    end;
end;

procedure TProxyForm.OnReaderErrorHandler(Reader: TReader; const Message: string;
    var Handled: Boolean);
begin
   ShowMessage(Message);
   Handled:= True;
end;

end.


